@@ -10,6 +10,7 @@
 #include <stdlib.h>
 #include <stdio.h>
 #include <unistd.h>
+#include <string.h>
 
 #include "list.h"
 #include "md5.h"
@@ -63,17 +64,23 @@ struct io_u {
 	struct timeval start_time;
 	struct timeval issue_time;
 
-	char *buf;
+	void *buf;
 	unsigned int buflen;
 	unsigned long long offset;
-	unsigned int index;
 
 	unsigned int resid;
 	unsigned int error;
 
-	unsigned char seen;
 	unsigned char ddir;
 
+	/*
+	 * io engine private data
+	 */
+	union {
+		unsigned int index;
+		unsigned int seen;
+	};
+
 	struct fio_file *file;
 
 	struct list_head list;
@@ -107,6 +114,7 @@ struct group_run_stats {
 enum fio_ddir {
 	DDIR_READ = 0,
 	DDIR_WRITE,
+	DDIR_SYNC,
 };
 
 /*
@@ -151,16 +159,15 @@ struct fio_file {
 
 	unsigned long *file_map;
 	unsigned int num_maps;
-
-	int fileno;
 };
 
 /*
  * This describes a single thread/process executing a fio job.
  */
 struct thread_data {
-	char name[32];
+	char *name;
 	char *directory;
+	char *filename;
 	char verror[80];
 	pthread_t thread;
 	int thread_number;
@@ -168,6 +175,7 @@ struct thread_data {
 	enum fio_filetype filetype;
 	struct fio_file *files;
 	unsigned int nr_files;
+	unsigned int nr_uniq_files;
 	unsigned int next_file;
 	int error;
 	pid_t pid;
@@ -178,26 +186,30 @@ struct thread_data {
 	enum fio_ddir ddir;
 	unsigned int iomix;
 	unsigned int ioprio;
+	unsigned int last_was_sync;
 
-	unsigned char sequential;
-	unsigned char odirect;
-	unsigned char invalidate_cache;
-	unsigned char create_serialize;
-	unsigned char create_fsync;
-	unsigned char end_fsync;
-	unsigned char sync_io;
-	unsigned char verify;
-	unsigned char use_thread;
-	unsigned char unlink;
-	unsigned char do_disk_util;
-	unsigned char override_sync;
-	unsigned char rand_repeatable;
-	unsigned char write_lat_log;
-	unsigned char write_bw_log;
+	unsigned int sequential;
+	unsigned int odirect;
+	unsigned int invalidate_cache;
+	unsigned int create_serialize;
+	unsigned int create_fsync;
+	unsigned int end_fsync;
+	unsigned int sync_io;
+	unsigned int verify;
+	unsigned int use_thread;
+	unsigned int unlink;
+	unsigned int do_disk_util;
+	unsigned int override_sync;
+	unsigned int rand_repeatable;
+	unsigned int write_lat_log;
+	unsigned int write_bw_log;
+	unsigned int norandommap;
+	unsigned int bs_unaligned;
 
-	unsigned int bs;
-	unsigned int min_bs;
-	unsigned int max_bs;
+	unsigned int bs[2];
+	unsigned int min_bs[2];
+	unsigned int max_bs[2];
+	unsigned int rw_min_bs;
 	unsigned int thinktime;
 	unsigned int fsync_blocks;
 	unsigned int start_delay;
@@ -214,12 +226,13 @@ struct thread_data {
 	os_cpu_mask_t cpumask;
 	unsigned int iolog;
 	unsigned int read_iolog;
-	unsigned int write_iolog;
 	unsigned int rwmixcycle;
 	unsigned int rwmixread;
+	unsigned int rwmixwrite;
 	unsigned int nice;
 
-	char *iolog_file;
+	char *read_iolog_file;
+	char *write_iolog_file;
 	void *iolog_buf;
 	FILE *iolog_f;
 
@@ -353,12 +366,24 @@ extern struct thread_data *threads;
 #define td_rw(td)		((td)->iomix != 0)
 
 #define BLOCKS_PER_MAP		(8 * sizeof(long))
-#define TO_MAP_BLOCK(td, f, b)	((b) - ((f)->file_offset / (td)->min_bs))
+#define TO_MAP_BLOCK(td, f, b)	((b) - ((f)->file_offset / (td)->rw_min_bs))
 #define RAND_MAP_IDX(td, f, b)	(TO_MAP_BLOCK(td, f, b) / BLOCKS_PER_MAP)
 #define RAND_MAP_BIT(td, f, b)	(TO_MAP_BLOCK(td, f, b) & (BLOCKS_PER_MAP - 1))
 
 #define MAX_JOBS	(1024)
 
+static inline int should_fsync(struct thread_data *td)
+{
+	if (td->last_was_sync)
+		return 0;
+	if (td->odirect)
+		return 0;
+	if (td_write(td) || td_rw(td) || td->override_sync)
+		return 1;
+
+	return 0;
+}
+
 struct disk_util_stat {
 	unsigned ios[2];
 	unsigned merges[2];
@@ -394,6 +419,9 @@ struct io_completion_data {
 #ifndef min
 #define min(a, b)	((a) < (b) ? (a) : (b))
 #endif
+#ifndef max
+#define max(a, b)	((a) > (b) ? (a) : (b))
+#endif
 
 /*
  * Log exports
@@ -431,7 +459,8 @@ extern unsigned long time_since_now(struct timeval *);
 extern unsigned long mtime_since_genesis(void);
 extern void __usec_sleep(unsigned int);
 extern void usec_sleep(struct thread_data *, unsigned long);
-extern void rate_throttle(struct thread_data *, unsigned long, unsigned int);
+extern void rate_throttle(struct thread_data *, unsigned long, unsigned int, int);
+extern void fill_start_time(struct timeval *);
 
 /*
  * Init functions
@@ -548,14 +577,13 @@ struct ioengine_ops {
 	struct io_u *(*event)(struct thread_data *, int);
 	int (*cancel)(struct thread_data *, struct io_u *);
 	void (*cleanup)(struct thread_data *);
-	int (*sync)(struct thread_data *, struct fio_file *);
 	void *data;
 	void *dlhandle;
 };
 
-#define FIO_IOOPS_VERSION	2
+#define FIO_IOOPS_VERSION	3
 
-extern struct ioengine_ops *load_ioengine(struct thread_data *, char *);
+extern struct ioengine_ops *load_ioengine(struct thread_data *, const char *);
 extern void close_ioengine(struct thread_data *);
 
 /*
